{
        
    ( - )

}
unit uPushPop;

interface

uses
  windows,Sysutils;
type


  PItemList=^TItemList;
  TItemList=record
    Prev,Next: PItemList;
    Data: PChar;
  end;

  TPushPop=class
  private
    FCS: RTL_CRITICAL_SECTION;
    FRootItem: PItemList;
    FLastItem: PItemList;
    FItemsCount: Integer;
    procedure Lock(L: Integer);
    procedure Unlock(L: Integer);
    procedure DeleteItem0;
    function GetCount: Integer;
  public
    BefEnter:      Integer;  //  ,      ()
    LstEnter:      Integer;
    LstLstEnter:   Integer;
    Entering:      Boolean;
    LstLeave:      Integer;
    LstLstLeave:   Integer;
    constructor Create;
    destructor Destroy;override;
    procedure Push(const Str: String);
    function Pop: String;
    function Peek: String;
    procedure Clear;
    property Count: Integer read GetCount;
  end;

implementation

{ TPushPop }

constructor TPushPop.Create;
begin
  New(FRootItem);
  New(FLastItem);

  FRootItem.Prev := nil;
  FRootItem.Next := FLastItem;
  FRootItem.Data := nil;
  FLastItem.Prev := FRootItem;
  FLastItem.Data := nil;
  FLastItem.Next := nil;

  FItemsCount := 0;
  InitializeCriticalSection(FCS);
end;

destructor TPushPop.Destroy;
begin
  Lock(1);
    while FRootItem.Next<>FLastItem do DeleteItem0;
    Dispose(FRootItem);
    Dispose(FLastItem);
  Unlock(-1);
end;

procedure TPushPop.Lock(L: Integer);
begin
  BefEnter:=L;
  Entering:=True;

  EnterCriticalSection(FCS);

  Entering:=False;
  LstLstEnter:=LstEnter;
  LstEnter:=L;
end;

procedure TPushPop.Unlock(L: Integer);
begin
  LeaveCriticalSection(FCS);

  LstLstLeave:=LstLeave;
  LstLeave:=L;
end;

procedure TPushPop.DeleteItem0;
var
  p: PItemList;
begin
  if FItemsCount=0 then Exit;
  p := FRootItem.Next;

  p.Prev.Next := p.Next;
  p.Next.Prev := p.Prev;

  Dispose(p);
  Dec(FItemsCount);
end;

function TPushPop.Peek: String;
begin
  Lock(2);
  try
    Result := String(FRootItem.Next.Data);
  finally
    Unlock(-2);
  end;
end;

function TPushPop.Pop: String;
begin
  Lock(3);
  try
    Result := String(FRootItem.Next.Data);
    FreeMem(FRootItem.Next.Data);
    DeleteItem0;
  finally
    Unlock(-3);
  end;
end;

procedure TPushPop.Push(const Str: String);
var
  p: PItemList;
  pi: PChar;
begin
  pi := AllocMem((Length(Str)+1)*SizeOf(Char));
  StrPCopy(pi,Str);
  Lock(4);
  try
    New(p);
    p.Data := pi;

    p.Prev := FLastItem.Prev;
    p.Next := FLastItem;
    FLastItem.Prev.Next := p;
    FLastItem.Prev := p;

    Inc(FItemsCount);
  finally
    Unlock(-4);
  end;
end;

function TPushPop.GetCount: Integer;
begin
  Lock(5);
  try
    Result := FItemsCount;
  finally
    Unlock(-5);
  end;
end;

procedure TPushPop.Clear;
begin
  while Count<>0 do Pop;
end;

end.
