{$WARN SYMBOL_PLATFORM OFF}
{
      .

     1.0.3    :

    1.             ,
       ..     TThreadLog.    tlWriteError, tlShiftError, tlErrorOpenLog
           20-         .    
           ,   ,    ,  10000.

    2.             ThrLogErr_yyyymmdd_hhmiss.txt
          ,    .
}

unit uThreadLog;

interface

uses windows, classes, SysUtils, uPushPop;

const

  MaxLogSizeConst = 8*1024*1024;  //     ()
  MaxLogNmbConst = 10;            //    

  MaxLenMsg = 500;                //    (.)

  ThrLogTimeout = 10000;          //    (.)      

  MaxWaitNmb = 20;                //        

type

  //     ,  

  TThreadLogError = (tlNone,                          // 
                     tlCreateObj,                     //  
                     tlWriteError,                    //    
                     tlShiftError,                    //     (  "")
                     tlInSuffSize,                    //   
                     tlErrorOpenLog,                  //   
                     tlErrorCreateDir,                //   
                     tlUnknow);                       // 
const

  ErrorMsg: array[tlNone..tlUnknow] of String=(
                      ' ',
                      '   ',
                      '    ',
                      '    ',
                      '    ',
                      '   ',
                      '   ',
                      ' ');

type
  //callback-   
  TProcErrorLog = procedure(const CodeError:TThreadLogError; const StrError: String) of object;

  TThreadLog = class(TThread)
  private
    FEvent: THandle;                         // Event   
    FActive: Boolean;                        //  
    FMaxLogSize: Integer;                    //     
    FMaxLogNmb: Integer;                     //    
    FLogName: String;                        //   
    FLogDir: String;                         //   
    FQueue: TPushPop;                        //    
    FError: TThreadLogError;                 //  
    FProcError: TProcErrorLog;               //    
    FResetTC: LongWord;                      //    (  )
    FThrEnded:    Boolean;                   //  ,    , ..  Execute  
    SpecErrToLog: Boolean;                   //  ,      
    SpecWaitCnt: Integer;                    // /     
    LastSpecError: TThreadLogError;          //  ,     
    procedure Init(const LogFileName: String; MaxLogSize: Integer; MaxLogNmb: Integer; Started: Boolean);
    function ChkLogDir: Boolean;             //    Log-
    function WriteMsg(FHandle: THandle): Boolean; //    
    function ShiftLogFiles: Boolean;              //   -  9->8->6...->1.   
    procedure Flush;                              //       
    procedure SetLastError(const Value: TThreadLogError);
    function IsSpecialError: Boolean;             //   ,       
  protected
    procedure Execute; override;
  public
    constructor Create(const LogFileName: String; MaxLogSize: Integer = 0; MaxLogNmb: Integer = 0; Started: Boolean = False);
    constructor CreateDefault;
    destructor Destroy; override;

    procedure Add(const s: String); overload;     //       
    procedure Add(const sl: TStringList); overload;
    procedure Start;                              //  
    procedure Stop;                               //  
    function  GetLockStr: String;                 //     
    property OnError: TProcErrorLog read FProcError write FProcError;
    property LastError: TThreadLogError read FError write SetLastError;
    property ThrEnded: Boolean read FThrEnded;
    property ResetTC: LongWord read FResetTC;
  end;

var

 ThrLog: TThreadLog;

implementation

{ TThreadLog }

procedure TThreadLog.Init(const LogFileName: String; MaxLogSize: Integer; MaxLogNmb: Integer; Started: Boolean);
begin
  FreeOnTerminate := False;

  if MaxLogSize = 0 then FMaxLogSize := MaxLogSizeConst
  else                   FMaxLogSize := MaxLogSize;

  if MaxLogNmb = 0 then FMaxLogNmb := MaxLogNmbConst
  else                  FMaxLogNmb := MaxLogNmb;

  FLogDir :=  ExtractFilePath(LogFileName);
  FLogName := ExtractFileName(LogFileName);
  FActive := False;
  FQueue := TPushPop.Create;
  FEvent := CreateEvent(nil,False,False,nil);
  if FEvent=0 then
  begin
    LastError := tlCreateObj;
  end
  else
  begin
    LastError := tlNone;
    ChkLogDir;
  end;
  Resume;
  if Started then Start;
end;

constructor TThreadLog.CreateDefault;
begin
  inherited Create(True);
  Init(ExtractFilePath(ParamStr(0))+'Logs\',0,0,False);
end;


constructor TThreadLog.Create(const LogFileName: String; MaxLogSize: Integer = 0; MaxLogNmb: Integer = 0; Started: Boolean = False);
begin
  inherited Create(True);
  Priority := tpHigher;  //    (  )         
  Init(LogFileName, MaxLogSize, MaxLogNmb, Started);
end;

destructor TThreadLog.Destroy;
begin
  FQueue.Clear;
  FQueue.Free;
  CloseHandle(FEvent);
  inherited;
end;

procedure TThreadLog.Execute;
var RC: DWORD;
  procedure ToErrLog;
  var TF: TextFile; DT: TDateTime; YY,MM,DD,HH,MI,SS,MS: Word;
  begin
    try
      DT := Now;
      DecodeDate(DT,YY,MM,DD); DecodeTime(DT,HH,MI,SS,MS);
      Assign(TF,FLogDir + 'ThrLogErr_' +
                IntToStr(YY) + FormatFloat('00',MM) + FormatFloat('00',DD) + '_' +
                FormatFloat('00',HH) + FormatFloat('00',MI) + FormatFloat('00',SS) + '.txt');
      Rewrite(TF);
      Writeln(TF,IntToStr(Ord(FError)));
      CloseFile(TF);
    except
    end;
  end;
begin
  if FError <> tlNone then Exit;
  SetEvent(FEvent);
  while not Terminated do
  begin
    RC := WaitForSingleObject(FEvent,1000);
    if IsSpecialError then    //   
      if SpecErrToLog then      //         
      begin
         ToErrLog;
         SpecErrToLog := False  //        
      end
      else                   //          
    else                     //    
      if FError <> tlNone then  //      ,
      begin                     //    !
        ToErrLog;
        Exit;
      end;

    case RC of
      WAIT_OBJECT_0, WAIT_TIMEOUT:
      begin
        FResetTC := GetTickCount;  //   ,  -     
        if SpecWaitCnt = 0 then
          Flush
        else
          Dec(SpecWaitCnt);        //     ,     Flush
      end;
    else
      begin
        LastError := tlUnknow;
        ToErrLog;
        Exit;
      end;
    end;
  end;
  FThrEnded := True;  // ,        
end;

//   -  9->8->6...->1.   
//  -   -
function TThreadLog.ShiftLogFiles: Boolean;
var NewName: String; ExtName: string; ExtPos: integer; i: integer; F: TSearchRec;
    FHandle: THandle;
begin
  Result:=True;
  try
    ExtName := ExtractfileExt(FLogName); //   
    ExtPos := Pos(ExtName, FLogName);
    NewName := FLogName;
    Delete(NewName, ExtPos + 1, Length(NewName) - ExtPos ) ; //  (. )
    NewName := FLogDir + NewName;

    //    -  
    for i := 0 to FMaxLogNmb - 2 do
      if FindFirst(NewName + IntToStr(i), faAnyFile, F) <> 0 then
        begin
          FHandle :=FileCreate(NewName + IntToStr(i));
          CloseHandle(FHandle);
        end
      else
        FindClose(F);

    //  ()   
    if Not DeleteFile(NewName + IntToStr(0)) then
    begin
     LastError := tlShiftError;
     Result:=False;
     Exit;
    end;

    for i:= 1 to FMaxLogNmb - 2 do
      if Not RenameFile(NewName + IntToStr(i),  NewName + IntToStr(i - 1)) then
      begin
       LastError := tlShiftError;
       Result:=False;
       Exit;
      end;

    //          
    if Not RenameFile(FLogDir + FLogName,  NewName + IntToStr(FMaxLogNmb - 2)) then
    begin
     LastError := tlShiftError;
     Result:=False;
     Exit;
    end;

  except
    Result:=False;
    LastError := tlShiftError;
  end;
end;

procedure TThreadLog.Flush;
var FHandle: THandle; Sz,SizeHigh: Integer;
begin
  if FActive and
     ((FError = tlNone) or
      IsSpecialError) then                                //    (  )     
  begin
    try
      FHandle := CreateFile(                              //  ,  
                 PChar(String(FLogDir + FLogName)),
                 GENERIC_READ + GENERIC_WRITE,
                 FILE_SHARE_READ,
                 nil,
                 OPEN_ALWAYS,
                 0,
                 0);
      if FHandle=INVALID_HANDLE_VALUE then
      begin
        LastError := tlErrorOpenLog;
        Exit;
      end;

      Sz := GetFileSize(FHandle,@SizeHigh);               //!!    
      if Sz+(SizeHigh shl 32) > FMaxLogSize then
        begin
          CloseHandle(FHandle);                           // ,  
          if not ShiftLogFiles then
          begin
            Exit; //  
          end
        else
          begin
            FHandle := CreateFile(                        //   ,  
                       PChar(String(FLogDir + FLogName)),
                       GENERIC_READ + GENERIC_WRITE,
                       FILE_SHARE_READ,
                       nil,
                       OPEN_ALWAYS,
                       0,
                       0);
            if FHandle=INVALID_HANDLE_VALUE then
            begin
              LastError := tlErrorOpenLog;
              Exit;
            end;
          end;
        end;

      try
        SetFilePointer(FHandle,0,nil,FILE_END);

        while FActive and
              ((FError = tlNone) or IsSpecialError) and
              (FQueue.Count > 0) do
          WriteMsg(FHandle);
      finally
        CloseHandle(FHandle);
      end;

      LastError := tlNone;       // ..   ,    
    except
      LastError := tlWriteError; //   
    end;
  end;
end;

function TThreadLog.ChkLogDir: Boolean;
begin
  Result := False;
  if not DirectoryExists(FLogDir) then
  begin
    if not ForceDirectories(FLogDir) then
    begin
      LastError := tlErrorCreateDir;
      Exit;
    end;
  end;

  Result := True;
end;

procedure TThreadLog.Add(const s: String);
var Msg: String;
begin
  Msg := s;
  if Length(Msg) > MaxLenMsg then SetLength(Msg,MaxLenMsg);
  FQueue.Push(Msg);
  if FActive then SetEvent(FEvent);
end;

procedure TThreadLog.Add(const sl: TStringList);  //    
var i: Integer; Msg: String;
begin
  for i:=1 to sl.Count do
  begin
    Msg := sl[i-1];
    if Length(Msg) > MaxLenMsg then SetLength(Msg,MaxLenMsg);
    FQueue.Push(Msg);
  end;
  if FActive then SetEvent(FEvent);
end;

function TThreadLog.WriteMsg(FHandle: THandle): Boolean;
var
  ResultCount:Cardinal;
  Len:Integer;
  Msg: AnsiString;
begin
  Result := True;
  if FQueue.Count=0 then Exit;
  Msg := FQueue.Peek+#13#10;
  Len := Length(Msg);
  if WriteFile(FHandle,Msg[1],Len,ResultCount,nil) then
  begin
    FQueue.Pop;
  end
  else
  begin
    if GetLastError=112
      then LastError := tlInSuffSize
      else LastError := tlWriteError;
    Result := False;
    Exit;
  end;
end;

procedure TThreadLog.Stop;
var Cnt: Integer;
begin
  while FActive and (LastError = tlNone) and Not(Terminated) and
        (FQueue.Count > 0) do Sleep(10);                   //       
  Sleep(10);

  FActive := False;
  Terminate;
  Cnt := 0;
  repeat
    Sleep(10); Inc(Cnt);
  until Terminated or (Cnt = 200);
end;

procedure TThreadLog.Start;
begin
  FActive := True;
  SetEvent(FEvent);
end;

procedure TThreadLog.SetLastError(const Value: TThreadLogError);
begin
  FError := Value;
  if IsSpecialError then
  begin
    SpecWaitCnt := MaxWaitNmb;                //   /  
    SpecErrToLog := LastSpecError <> FError;  //  ,      ,  ,    
    LastSpecError := FError;                  //        
  end
  else
    if FError <> tlNone then
      FActive := False
    else
      LastSpecError := tlNone;                //    
end;

//   ,       
function TThreadLog.IsSpecialError;
begin
  Result := (FError in [tlErrorOpenLog, tlShiftError, tlWriteError]) and
            (FQueue.Count < 10000);
end;


function  TThreadLog.GetLockStr: String;  //    (  )
var S: String;
begin
  if FQueue.Entering then S := '!!! ENTERING' else S := 'NoEntering';
  S := #13#10 +
       S + ', LstEnter - ' + IntToStr(FQueue.LstEnter) +
           ', BefEnter - ' + IntToStr(FQueue.BefEnter) +
           ', LstLstEnter - ' + IntToStr(FQueue.LstLstEnter) + ',' +
           #13#10'            ' +
           'LstLeave - ' + IntToStr(FQueue.LstLeave) +
           ', LstLstLeave - ' + IntToStr(FQueue.LstLstLeave) + #13#10#13#10;
  Result := S;
end;

initialization
  ThrLog := nil;
end.
